@@ -91,9 +91,9 @@ enum {
   SILCTXT_PUBKEY_NO_MATCH,
   SILCTXT_PUBKEY_MAYBE_EXPIRED,
   SILCTXT_PUBKEY_MITM_ATTACK,
-  SILCTXT_GETKEY_NOKEY,
-  SILCTXT_GETKEY_VERIFIED,
-  SILCTXT_GETKEY_DISCARD,
+  SILCTXT_PUBKEY_NOKEY,
+  SILCTXT_PUBKEY_VERIFIED,
+  SILCTXT_PUBKEY_NOTVERIFIED,
 
   SILCTXT_FILL_4,
 
@@ -122,6 +122,7 @@ enum {
   SILCTXT_WATCH_UMODE_CHANGE,
   SILCTXT_WATCH_NICK_CHANGE,
   SILCTXT_MESSAGE_DATA,
+  SILCTXT_STATS,
 
   SILCTXT_FILL_5,
 
@@ -142,6 +143,34 @@ enum {
   SILCTXT_FILE_ERROR_NO_SUCH_FILE,
   SILCTXT_FILE_ERROR_PERMISSION_DENIED,
   SILCTXT_FILE_CLOSED,
+
+  SILCTXT_FILL_6,
+
+  SILCTXT_ATTR_HEADER,
+  SILCTXT_ATTR_FOOTER,
+  SILCTXT_ATTR_ALLOW,
+  SILCTXT_ATTR_VCARD_FILE,
+  SILCTXT_ATTR_SERVICES,
+  SILCTXT_ATTR_SERVICE,
+  SILCTXT_ATTR_STATUS_MOOD,
+  SILCTXT_ATTR_STATUS_TEXT,
+  SILCTXT_ATTR_STATUS_MESSAGE,
+  SILCTXT_ATTR_STATUS_MESSAGE_FILE,
+  SILCTXT_ATTR_PREFERRED_LANGUAGE,
+  SILCTXT_ATTR_PREFERRED_CONTACT,
+  SILCTXT_ATTR_GEOLOCATION,
+  SILCTXT_ATTR_DEVICE_INFO,
+  SILCTXT_ATTR_PUBLIC_KEYS,
+  SILCTXT_ATTR_TIMEZONE,
+  SILCTXT_ATTR_TIMEZONE_ALLOW,
+  SILCTXT_ATTR_USER_SIGN_VERIFIED,
+  SILCTXT_ATTR_USER_SIGN_FAILED,
+  SILCTXT_ATTR_USER_SIGN_NOT_PRESENT,
+  SILCTXT_ATTR_SERVER_SIGN_VERIFIED,
+  SILCTXT_ATTR_SERVER_SIGN_FAILED,
+  SILCTXT_ATTR_EXTENSION,
+  SILCTXT_ATTR_SAVE,
+  SILCTXT_ATTR_SAVED,
 };
 
 extern FORMAT_REC fecommon_silc_formats[];
