@@ -17,7 +17,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: client_prvmsg.c,v 1.38 2002/05/28 11:56:23 priikone Exp $ */
+/* $Id: client_prvmsg.c,v 1.40 2002/10/14 19:04:36 priikone Exp $ */
 /* This file includes the private message sending and receiving routines
    and private message key handling routines. */
 
@@ -173,7 +173,7 @@ void silc_client_private_message(SilcClient client,
     }
 
     /* Resolve the client info */
-    silc_client_get_client_by_id_resolve(client, conn, remote_id,
+    silc_client_get_client_by_id_resolve(client, conn, remote_id, NULL,
 					 silc_client_private_message_cb,
 					 silc_packet_context_dup(packet));
     return;
@@ -287,6 +287,7 @@ void silc_client_private_message_key(SilcClient client,
     return;
 
   silc_client_get_client_by_id_resolve(client, sock->user_data, remote_id,
+				       NULL,
 				       silc_client_private_message_key_cb,
 				       silc_packet_context_dup(packet));
   silc_free(remote_id);
@@ -356,7 +357,7 @@ int silc_client_add_private_message_key(SilcClient client,
   /* Produce the key material as the protocol defines */
   keymat = silc_calloc(1, sizeof(*keymat));
   if (silc_ske_process_key_material_data(key, key_len, 16, 256, 16, 
-					 client->internal->md5hash, keymat) 
+					 client->md5hash, keymat) 
       != SILC_SKE_STATUS_OK)
     return FALSE;
 
